/* $Id: cpl_mask-test.c,v 1.38 2009/05/18 10:01:05 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2008 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2009/05/18 10:01:05 $
 * $Revision: 1.38 $
 * $Name:  $
 */

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>

#include "cpl_mask.h"
#include "cpl_image_gen.h"
#include "cpl_stats.h"
#include "cpl_image_io.h"
#include "cpl_memory.h"
#include "cpl_msg.h"
#include "cpl_test.h"

/*-----------------------------------------------------------------------------
                                  Defines
 -----------------------------------------------------------------------------*/

#ifndef IMAGESZ
#define IMAGESZ 512
#endif

#ifndef MASKSZ
#define MASKSZ 8
#endif

/*-----------------------------------------------------------------------------
                                  Private functions
 -----------------------------------------------------------------------------*/

static cpl_error_code cpl_mask_shift_bf(cpl_mask *, int, int);
static void cpl_mask_shift_test(int, int);
static void cpl_mask_test(cpl_boolean);

static double tshift   = 0.0;
static double tshiftbf = 0.0;

/*-----------------------------------------------------------------------------
                                  Main
 -----------------------------------------------------------------------------*/
int main(void)
{
    cpl_image       *   imd;
    double              median,
                        med_dist;
    cpl_mask        *   bin1,
                    *   bin2,
                    *   bin1_dup;
    cpl_mask        *   null;
    cpl_image       *   labels1,
                    *   labels2;
    double              threshold;
    int                 count;
    int                 nb_objs1,
                        nb_objs2;
    cpl_matrix      *   kernel;
    double          *   ker_arr;
    const int           new_pos[] = {1,2,3,4};
    const int           nsize = IMAGESZ;
    cpl_binary      *   bins;
    cpl_error_code      error;
    cpl_boolean         do_bench;
    FILE            *   stream;

    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    do_bench = cpl_msg_get_level() <= CPL_MSG_INFO ? CPL_TRUE : CPL_FALSE;

    stream = cpl_msg_get_level() > CPL_MSG_INFO
        ? fopen("/dev/null", "a") : stdout;

    /* Insert tests below */

    cpl_test_nonnull( stream );

    cpl_mask_test(do_bench);

    cpl_mask_shift_test(1, 1);
    cpl_mask_shift_test(1, 2);
    cpl_mask_shift_test(2, 1);
    cpl_mask_shift_test(2, 2);
    cpl_mask_shift_test(MASKSZ, 1);
    cpl_mask_shift_test(1, MASKSZ);
    cpl_mask_shift_test(MASKSZ, 2);
    cpl_mask_shift_test(2, MASKSZ);
    cpl_mask_shift_test(MASKSZ, 3);
    cpl_mask_shift_test(3, MASKSZ);
    cpl_mask_shift_test(MASKSZ, MASKSZ);

    if (do_bench) {
        cpl_mask_shift_test(IMAGESZ, MASKSZ);
    }

    cpl_msg_info(cpl_func, "Time to shift [s]: %g <=> %g (brute force)",
                 tshift, tshiftbf);

    /* Create test image */
    cpl_msg_info("","Create a %dx%d test image", nsize, nsize);
    imd = cpl_image_fill_test_create(nsize, nsize);
    cpl_test_nonnull(imd);
    
    /* Compute the median and the av. dist. to the median of the loaded image */
    median = cpl_image_get_median_dev(imd, &med_dist);

    /* Binarize the image with a first threshold */
    cpl_msg_info("","Threshold the test image with median+2*sigma -> bin1");
    threshold = median + 2.0 * med_dist;
    bin1 = cpl_mask_threshold_image_create(imd, threshold, DBL_MAX);
    cpl_test_nonnull(bin1);

    cpl_test_zero(cpl_mask_is_empty(bin1));

    /* Test wrap / unwrap */
    bins = cpl_calloc(100, sizeof(cpl_binary));
    bin2 = cpl_mask_wrap(10, 10, bins);
    cpl_test(cpl_mask_is_empty(bin2));

    cpl_mask_delete(bin2);
    bins = cpl_calloc(100, sizeof(cpl_binary));
    bin2 = cpl_mask_wrap(10, 10, bins);
    cpl_test(cpl_mask_is_empty(bin2));
    cpl_test(bins == cpl_mask_unwrap(bin2));
    cpl_free(bins);

    /* Binarize the image with a second threshold */
    cpl_msg_info("","Threshold the test image with median+5*sigma -> bin2");
    threshold = median + 5.0 * med_dist; 
    bin2 = cpl_mask_threshold_image_create(imd, threshold, DBL_MAX);
    cpl_test_nonnull(bin2);
    cpl_image_delete(imd);
   
    /* Test cpl_mask_get() */
    /* Test cpl_mask_set() */
    count = cpl_mask_count(bin1);
    if (cpl_mask_get(bin1, 1, 1) == CPL_BINARY_0) {
        cpl_test_zero(cpl_mask_count_window(bin1, 1, 1, 1, 1));
        cpl_mask_set(bin1, 1, 1, CPL_BINARY_1);
        cpl_test_eq(cpl_mask_count(bin1), count + 1);
        cpl_test_eq(cpl_mask_count_window(bin1, 1, 1, 1, 1), 1);
    } else {
        cpl_test_eq(cpl_mask_count_window(bin1, 1, 1, 1, 1), 1);
        cpl_mask_set(bin1, 1, 1, CPL_BINARY_0);
        cpl_test_eq(cpl_mask_count(bin1), count - 1);
        cpl_test_zero(cpl_mask_count_window(bin1, 1, 1, 1, 1));
        /* Later a non-zero mask is required */
        cpl_mask_set(bin1, 1, 1, CPL_BINARY_1);
    }

    /* Test error handling of cpl_mask_set() */
    error = cpl_mask_set(NULL, 1, 1, CPL_BINARY_1);
    cpl_test_eq_error(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_set(bin1, 0, 1, CPL_BINARY_1);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1, 0, CPL_BINARY_1);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1, 1+nsize, CPL_BINARY_1);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1+nsize, 1, CPL_BINARY_1);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_set(bin1, 1, 1, (cpl_binary)-1);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    /* Compute number of selected pixels */
    cpl_test_eq(cpl_mask_count(bin1),
                cpl_mask_count_window(bin1, 1, 1, nsize, nsize));
    
    /* Test cpl_mask_duplicate() */
    bin1_dup = cpl_mask_duplicate(bin1);
    cpl_test_eq_mask(bin1, bin1_dup);
    cpl_test_eq(cpl_mask_is_empty(bin1), cpl_mask_is_empty(bin1_dup));
    cpl_test_eq(cpl_mask_count(bin1),    cpl_mask_count(bin1_dup));

    /* Test error handling of cpl_mask_collapse_create() */
    null = cpl_mask_collapse_create(NULL, 0);
    cpl_test_error(CPL_ERROR_NULL_INPUT);
    cpl_test_null(null);

    null = cpl_mask_collapse_create(bin1, 2);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_null(null);

    null = cpl_mask_collapse_create(bin1, -1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_null(null);


    /* Test cpl_mask_turn() */
    /* Rotate the mask all the way around */
    cpl_test(cpl_mask_turn(bin1,   9) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1,   6) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1, -13) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1,  10) == CPL_ERROR_NONE);
    cpl_test(cpl_mask_turn(bin1,   4) == CPL_ERROR_NONE);
    cpl_test_eq_mask(bin1, bin1_dup);
    /* Compare to its copy */
    cpl_test_zero(cpl_mask_xor(bin1_dup, bin1));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_zero(cpl_mask_count(bin1_dup));
    cpl_test_eq(cpl_mask_is_empty(bin1_dup), CPL_TRUE);
    cpl_mask_delete(bin1_dup);

    /* Build a 3x3 kernel */
    ker_arr = cpl_malloc(9*sizeof(double));
    ker_arr[0] = 1;
    ker_arr[1] = 1;
    ker_arr[2] = 1;
    ker_arr[3] = 1;
    ker_arr[4] = 1;
    ker_arr[5] = 1;
    ker_arr[6] = 1;
    ker_arr[7] = 1;
    ker_arr[8] = 1;
    kernel = cpl_matrix_wrap(3, 3, ker_arr);

    cpl_msg_info("","Apply a morphological closing on bin1");
    /* Test erosion function */
    cpl_test_zero(cpl_mask_erosion(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    /* Test dilation function */
    cpl_test_zero(cpl_mask_dilation(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_msg_info("","Apply a morphological opening on bin1");
    /* Test opening function */
    cpl_test_zero(cpl_mask_opening(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);
 
    cpl_msg_info("","Apply a morphological closing on bin2");
    /* Test closing function */
    cpl_test_zero(cpl_mask_closing(bin1, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_msg_info("","Apply a morphological opening on bin2");
    /* Test dilation function */
    cpl_test_zero(cpl_mask_dilation(bin2, kernel));
    cpl_test_error(CPL_ERROR_NONE);


    /* Test erosion function */
    cpl_test_zero(cpl_mask_erosion(bin2, kernel));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_matrix_delete(kernel);

    /* Labelize the binary maps */
    labels1 = cpl_image_labelise_mask_create(bin1, &nb_objs1);
    cpl_test_nonnull(labels1);

    cpl_msg_info("","Labelize bin1 ---> %d objects found", nb_objs1);
    labels2 = cpl_image_labelise_mask_create(bin2, &nb_objs2);
    cpl_test_nonnull(labels2);

    cpl_msg_info("","Labelize bin2 ---> %d objects found", nb_objs2);
   
    /* sprintf(outname, "labelized1.fits"); */
    /* cpl_image_save(labels1, outname, CPL_BPP_IEEE_FLOAT, NULL,
                      CPL_IO_CREATE); */
    /* sprintf(outname, "labelized2.fits"); */
    /* cpl_image_save(labels2, outname, CPL_BPP_IEEE_FLOAT, NULL,
                      CPL_IO_CREATE); */
    
    cpl_image_delete(labels1);
    cpl_image_delete(labels2);

    /* The following tests need to be done on a non-empty mask */ 
    cpl_test_zero(cpl_mask_is_empty(bin2));

    cpl_test_zero(cpl_mask_set(bin1, IMAGESZ, IMAGESZ, CPL_BINARY_1));
    cpl_test_zero(cpl_mask_set(bin2, IMAGESZ, IMAGESZ, CPL_BINARY_0));

    /* Test cpl_mask_and */
    cpl_test_zero(cpl_mask_and(bin1, bin2));
    cpl_test_error(CPL_ERROR_NONE);
    cpl_test_leq(cpl_mask_count(bin1), cpl_mask_count(bin2));
    
    /* Test cpl_mask_or */
    cpl_test_zero(cpl_mask_or(bin1, bin2));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_eq_mask(bin1, bin2);
     
    /* Test cpl_mask_not */
    cpl_test_zero(cpl_mask_not(bin2));
    cpl_test_error(CPL_ERROR_NONE);

    /* Test cpl_mask_xor */
    cpl_test_zero(cpl_mask_xor(bin1, bin2));
    cpl_test_error(CPL_ERROR_NONE);

    /* bin1 is now all ones... */

    /* Test cpl_mask_copy() */
    cpl_test_zero(cpl_mask_copy(bin2, bin1, 1, 1));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test_eq_mask(bin1, bin2);

    cpl_test_zero(cpl_mask_not(bin2));
    cpl_test_error(CPL_ERROR_NONE);

    cpl_test(cpl_mask_is_empty(bin2));

    /* Test error handling in cpl_mask_copy() */
    error = cpl_mask_copy(NULL, bin2, 1, 1);
    cpl_test_error(CPL_ERROR_NULL_INPUT);
    cpl_test_eq(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_copy(bin1, NULL, 1, 1);
    cpl_test_error(CPL_ERROR_NULL_INPUT);
    cpl_test_eq(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_copy(bin2, bin1, 0, 1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_copy(bin2, bin1, 1, 0);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_copy(bin2, bin1, nsize+1, 1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_copy(bin2, bin1, 1, nsize+1);
    cpl_test_error(CPL_ERROR_ILLEGAL_INPUT);
    cpl_test_eq(error, CPL_ERROR_ILLEGAL_INPUT);

    cpl_mask_delete(bin2);
    
    /* Test cpl_mask_extract() */
    bin2 = cpl_mask_extract(bin1, 1, 1, 1, 1);
    cpl_test_nonnull(bin2);
    cpl_test_eq(cpl_mask_get_size_x(bin2), 1);
    cpl_test_eq(cpl_mask_get_size_y(bin2), 1);
    cpl_mask_delete(bin2);
    
    /* Test cpl_mask_shift() */
    cpl_test_zero(cpl_mask_shift(bin1, 1, 1));
    cpl_test_error(CPL_ERROR_NONE);
    
    cpl_test_zero(cpl_mask_set(bin1, IMAGESZ, IMAGESZ, CPL_BINARY_1));

    cpl_test_zero(cpl_mask_is_empty(bin1));

    if (IMAGESZ % 2 == 0) {
        /* Test cpl_mask_move() */
        cpl_test_zero(cpl_mask_move(bin1, 2, new_pos));
        cpl_test_error(CPL_ERROR_NONE);
    }

    /* Test cpl_mask_extract_subsample() */
    bin2 = cpl_mask_extract_subsample(bin1, 2, 2);
    cpl_test_nonnull(bin2);

    cpl_test_zero(cpl_mask_dump_window(bin1, 1, 1, MASKSZ/2, MASKSZ/2, stream));
    cpl_test_zero(cpl_mask_dump_window(bin2, 1, 1, 1, 1, stream));

    error = cpl_mask_dump_window(bin2, 1, 1, IMAGESZ, MASKSZ, stream);
    cpl_test_eq_error(error, CPL_ERROR_ACCESS_OUT_OF_RANGE);
    error = cpl_mask_dump_window(bin2, 1, 1, MASKSZ, IMAGESZ, stream);
    cpl_test_eq_error(error, CPL_ERROR_ACCESS_OUT_OF_RANGE);
    error = cpl_mask_dump_window(bin2, 0, 1, MASKSZ, MASKSZ, stream);
    cpl_test_eq_error(error, CPL_ERROR_ACCESS_OUT_OF_RANGE);
    error = cpl_mask_dump_window(bin2, 1, 0, MASKSZ, MASKSZ, stream);
    cpl_test_eq_error(error, CPL_ERROR_ACCESS_OUT_OF_RANGE);

    error = cpl_mask_dump_window(NULL, 1, 1, MASKSZ, MASKSZ, stream);
    cpl_test_eq_error(error, CPL_ERROR_NULL_INPUT);

    error = cpl_mask_dump_window(bin1, 1, 1, MASKSZ, MASKSZ, NULL);
    cpl_test_eq_error(error, CPL_ERROR_NULL_INPUT);

    cpl_mask_delete(bin1);
    cpl_mask_delete(bin2);

    if (stream != stdout) cpl_test_zero( fclose(stream) );

    /* End of tests */
    return cpl_test_end(0);
}



/*----------------------------------------------------------------------------*/
/**
  @brief    Shift a mask using brute force
  @param    in      mask to shift
  @param    x_shift shift to apply in x
  @param    y_shift shift to apply in y
  @return   the #_cpl_error_code_ or CPL_ERROR_NONE
  @see      cpl_image_shift()

 */
/*----------------------------------------------------------------------------*/
static cpl_error_code cpl_mask_shift_bf(cpl_mask * in, 
                                        int        x_shift,
                                        int        y_shift)
{
    int                 xs_abs, ys_abs;
    cpl_mask        *   loc;
    cpl_binary      *   ploc;
    cpl_binary      *   pin;
    int                 i, j;
    const int nx = cpl_mask_get_size_x(in);
    const int ny = cpl_mask_get_size_y(in); 
    /* Initialise */
    xs_abs = abs(x_shift);
    ys_abs = abs(y_shift);
    
   /* Check entries */
    cpl_ensure_code(in != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code((xs_abs<nx)&&(ys_abs<ny), CPL_ERROR_ILLEGAL_INPUT);

    /* Handle the case where nothing has to be done */
    if (x_shift==0 && y_shift==0) return CPL_ERROR_NONE;

    /* Create the local mask */
    loc = cpl_mask_duplicate(in);
    ploc = cpl_mask_get_data(loc);
    pin = cpl_mask_get_data(in);

    for (j=0; j<ny; j++) {
        for (i=0; i<nx; i++) {
            if ((i-x_shift>=0) && (i-x_shift<nx) &&
                    (j-y_shift>=0) && (j-y_shift<ny)) {
                pin[i+j*nx] = ploc[(i-x_shift)+(j-y_shift)*nx];
            } else {
                pin[i+j*nx] = CPL_BINARY_1;
            }
        }
    }
    cpl_mask_delete(loc);
    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief  Test cpl_mask_shift()
  @param  m      The kernel X-size
  @param  n      The kernel Y-size
  @return void

 */
/*----------------------------------------------------------------------------*/
static void cpl_mask_shift_test(int m, int n)
{

    cpl_image * testimg = cpl_image_new(m, n, CPL_TYPE_FLOAT);
    cpl_mask  * mask0   = cpl_mask_new(m, n);
    cpl_error_code error;
    int i, j;

    cpl_msg_info(cpl_func, "Shift(%d,%d)", m, n);

    error = cpl_mask_shift(NULL, 0, 0);
    cpl_test_eq_error(error, CPL_ERROR_NULL_INPUT);


    error = cpl_mask_shift(mask0, m, 0);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_shift(mask0, -m, 0);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);


    error = cpl_mask_shift(mask0, 0, n);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    error = cpl_mask_shift(mask0, 0, -n);
    cpl_test_eq_error(error, CPL_ERROR_ILLEGAL_INPUT);

    if (m == 1 && n == 1) {
        cpl_mask * mask1 = cpl_mask_new(m, n);
        cpl_test_zero(cpl_mask_shift(mask1, 0, 0));
        cpl_mask_delete(mask1);
    } else {
        cpl_mask * mask1 = cpl_mask_new(m, n);
        cpl_mask * mask2 = cpl_mask_new(m, n);
        int count;

        do {
            /* Need a mask with a mix of elements */
            cpl_test_zero(cpl_image_fill_noise_uniform(testimg, 0.0, 1.0));
            cpl_mask_delete(mask0);
            mask0 = cpl_mask_threshold_image_create(testimg, 0.25, 0.75);
            count = cpl_mask_count(mask0);
        } while (count == 0 || count == m*n);

        for (j = 1-n; j < n; j++) {
            for (i = 1-m; i < m; i++) {
                double now;

                cpl_mask_copy(mask1, mask0, 1, 1);
                cpl_mask_copy(mask2, mask0, 1, 1);

                now = cpl_test_get_cputime();
                cpl_test_zero(cpl_mask_shift(mask1, i, j));
                tshift += cpl_test_get_cputime() - now;

                now = cpl_test_get_cputime();
                cpl_test_zero(cpl_mask_shift_bf(mask2, i, j));
                tshiftbf += cpl_test_get_cputime() - now;

                cpl_test_eq_mask(mask1, mask2);

            }
        }
        cpl_mask_delete(mask1);
        cpl_mask_delete(mask2);
    }

    cpl_image_delete(testimg);
    cpl_mask_delete(mask0);
}


/*----------------------------------------------------------------------------*/
/**
  @internal
  @param do_bench True iff benchmarking is enabled
  @brief          Test various cpl_mask functions
  @return         void
  
 */
/*----------------------------------------------------------------------------*/
static void cpl_mask_test(cpl_boolean do_bench)
{

    const int npix[] = {1, 2, 3, 5, 10, 11, do_bench ? 2*IMAGESZ : 13};

    const cpl_binary buf9[] = {CPL_BINARY_1, CPL_BINARY_0, CPL_BINARY_1,
                               CPL_BINARY_0, CPL_BINARY_0, CPL_BINARY_1,
                               CPL_BINARY_1, CPL_BINARY_0, CPL_BINARY_1};

    /* Rotation 90 degrees counterclockwise */
    const cpl_binary rot9[] = {CPL_BINARY_1, CPL_BINARY_1, CPL_BINARY_1,
                               CPL_BINARY_0, CPL_BINARY_0, CPL_BINARY_0,
                               CPL_BINARY_1, CPL_BINARY_0, CPL_BINARY_1};

    /* This mask is not modified while wrapping around this buffer */
    cpl_mask     * ref = cpl_mask_wrap(3, 3, (cpl_binary*)rot9);
    cpl_mask     * raw = cpl_mask_new(3, 3);
    cpl_error_code error;
    size_t         ipix;
    double         tflip = 0.0;
    double         tturn = 0.0;
    double         tassi = 0.0;
    double         tcol0 = 0.0;
    double         tcol1 = 0.0;

    /* Test 1: Verify direction of 90 degree rotation */
    cpl_test_nonnull(memcpy(cpl_mask_get_data(raw), buf9, 9*sizeof(cpl_binary)));

    cpl_test_zero(cpl_mask_turn(raw, 1));
    cpl_test_eq_mask(raw, ref);

    cpl_mask_delete(raw);
    (void)cpl_mask_unwrap(ref);

    /* Test 2: Check error handling */
    error = cpl_image_turn(NULL, 1);
    cpl_test_eq_error(error, CPL_ERROR_NULL_INPUT);

    /* Test 3: Verify consistency of rotations across types */
    for (ipix = 0; ipix < sizeof(npix)/sizeof(npix[0]); ipix++) {
        const int nx = npix[ipix];
        size_t jpix;
        for (jpix = 0; jpix < sizeof(npix)/sizeof(npix[0]); jpix++) {
            const int ny = npix[jpix];

            cpl_image * noise = cpl_image_new(nx, ny, CPL_TYPE_FLOAT);
            cpl_mask * collapsed;

            double ttemp;
            int i;

            cpl_test_zero(cpl_image_fill_noise_uniform(noise, (double)-ny,
                                                       (double)nx));

            /* About 1/4 of the pixels are flagged */
            raw = cpl_mask_threshold_image_create(noise, -0.25*ny, 0.25*nx);
            cpl_image_delete(noise);

            ref = cpl_mask_extract(raw, 1, 1, nx, ny);

#if 1
            ttemp = cpl_test_get_cputime();

            /* Turn the mask all the way around */
            cpl_test_zero(cpl_mask_turn(raw, 1));
            cpl_test_zero(cpl_mask_turn(raw, 2));
            cpl_test_zero(cpl_mask_turn(raw, 3));
            cpl_test_zero(cpl_mask_turn(raw, 4));
            cpl_test_zero(cpl_mask_turn(raw,-1));
            cpl_test_zero(cpl_mask_turn(raw,-2));
            cpl_test_zero(cpl_mask_turn(raw,-3));
            cpl_test_zero(cpl_mask_turn(raw, 0));

            tturn += cpl_test_get_cputime() - ttemp;

            cpl_test_eq_mask(raw, ref);

            ttemp = cpl_test_get_cputime();

            /* Flip the mask all the way around */
            cpl_test_zero(cpl_mask_flip(raw, 0));
            cpl_test_zero(cpl_mask_flip(raw, 1));
            cpl_test_zero(cpl_mask_flip(raw, 2));
            cpl_test_zero(cpl_mask_flip(raw, 3));
            cpl_test_zero(cpl_mask_flip(raw, 0));
            cpl_test_zero(cpl_mask_flip(raw, 1));
            cpl_test_zero(cpl_mask_flip(raw, 2));
            cpl_test_zero(cpl_mask_flip(raw, 3));

            tflip += cpl_test_get_cputime() - ttemp;

            cpl_test_eq_mask(raw, ref);

#endif

            ttemp = cpl_test_get_cputime();
            cpl_test_zero(cpl_mask_or(raw, ref));
            cpl_test_eq_mask(raw, ref);

            cpl_test_zero(cpl_mask_and(raw, ref));
            cpl_test_eq_mask(raw, ref);

            /* Empty the mask */
            cpl_test_zero(cpl_mask_xor(raw, ref));
            cpl_test(cpl_mask_is_empty(raw));

            /* cpl_mask_copy()... */
            cpl_test_nonnull(memcpy(cpl_mask_get_data(raw),
                                    cpl_mask_get_data_const(ref), 
                                   nx * ny * sizeof(cpl_binary)));

            /* Empty the mask */
            cpl_test_zero(cpl_mask_not(raw));
            cpl_test_zero(cpl_mask_and(raw, ref));
            cpl_test(cpl_mask_is_empty(raw));


            /* cpl_mask_copy()... */
            cpl_test_nonnull(memcpy(cpl_mask_get_data(raw),
                                    cpl_mask_get_data_const(ref),
                                    nx * ny * sizeof(cpl_binary)));

            /* Fill the mask */
            cpl_test_zero(cpl_mask_not(raw));
            cpl_test_zero(cpl_mask_or(raw, ref));
            cpl_test_eq(cpl_mask_count(raw), nx * ny);

            tassi += cpl_test_get_cputime() - ttemp;

            for (i = 1; i <= nx; i++) {
                /* First collapsed element will be CPL_BINARY_0 */
                cpl_mask_set(raw, i, 1, CPL_BINARY_0);
            }
            if (ny > 2) {
                /* Second collapsed element will be CPL_BINARY_0 */
                cpl_mask_set(raw, 1, 2, CPL_BINARY_0);
            }

            ttemp = cpl_test_get_cputime();

            collapsed = cpl_mask_collapse_create(raw, 1);
            tcol1 += cpl_test_get_cputime() - ttemp;

            cpl_test_error(CPL_ERROR_NONE);
            cpl_test_nonnull(collapsed);

            cpl_test_eq(cpl_mask_get_size_x(collapsed), 1);
            cpl_test_eq(cpl_mask_get_size_y(collapsed), ny);

            cpl_test_zero(cpl_mask_get(collapsed, 1, 1));
            if (ny > 2) {
                cpl_test_zero(cpl_mask_get(collapsed, 1, 2));
            }
            cpl_test_eq(cpl_mask_count(collapsed), ny > 2 ? ny - 2 : ny - 1);

            cpl_mask_delete(collapsed);

#if 1
            /* cpl_mask_copy()... */
            cpl_test_nonnull(memcpy(cpl_mask_get_data(raw),
                                    cpl_mask_get_data_const(ref),
                                    nx * ny * sizeof(cpl_binary)));


            /* Fill the mask */
            ttemp = cpl_test_get_cputime();
            cpl_test_zero(cpl_mask_not(raw));
            cpl_test_zero(cpl_mask_or(raw, ref));
            tassi += cpl_test_get_cputime() - ttemp;
            cpl_test_eq(cpl_mask_count(raw), nx * ny);

            for (i = 1; i <= ny; i++) {
                /* First collapsed element will be CPL_BINARY_0 */
                cpl_mask_set(raw, 1, i, CPL_BINARY_0);
            }
            if (nx > 2) {
                /* Second collapsed element will be CPL_BINARY_0 */
                cpl_mask_set(raw, 2, 1, CPL_BINARY_0);
            }

            ttemp = cpl_test_get_cputime();

            collapsed = cpl_mask_collapse_create(raw, 0);
            tcol0 += cpl_test_get_cputime() - ttemp;

            cpl_test_error(CPL_ERROR_NONE);
            cpl_test_nonnull(collapsed);

            cpl_test_eq(cpl_mask_get_size_x(collapsed), nx);
            cpl_test_eq(cpl_mask_get_size_y(collapsed), 1);

            cpl_test_zero(cpl_mask_get(collapsed, 1, 1));
            if (nx > 2) {
                cpl_test_zero(cpl_mask_get(collapsed, 2, 1));
            }
            cpl_test_eq(cpl_mask_count(collapsed), nx > 2 ? nx - 2 : nx - 1);

            cpl_mask_delete(collapsed);
#endif

            if (nx > 1 && ny > 1) {
                int nzero = 0;
                const int nmin = nx < ny ? nx : ny;

                /* Fill the mask */
                ttemp = cpl_test_get_cputime();
                cpl_test_zero(cpl_mask_xor(raw, raw));
                cpl_test(cpl_mask_is_empty(raw));
                cpl_test_zero(cpl_mask_not(raw));
                tassi += cpl_test_get_cputime() - ttemp;
                cpl_test_eq(cpl_mask_count(raw), nx * ny);

                /* Every other element on the diagonal is CPL_BINARY_0 */
                for (i = 1; i <= nmin; i += 2) {
                    cpl_test_zero(cpl_mask_set(raw, i, i, CPL_BINARY_0));
                    nzero++;
                }
                cpl_test_eq(cpl_mask_count(raw), nx * ny-nzero);

                ttemp = cpl_test_get_cputime();

                collapsed = cpl_mask_collapse_create(raw, 1);
                tcol1 += cpl_test_get_cputime() - ttemp;

                cpl_test_error(CPL_ERROR_NONE);
                cpl_test_nonnull(collapsed);

                cpl_test_eq(cpl_mask_get_size_x(collapsed), 1);
                cpl_test_eq(cpl_mask_get_size_y(collapsed), ny);

                for (i = 1; i <= nmin; i += 2) {
                    cpl_test_zero(cpl_mask_get(collapsed, 1, i));
                }

                cpl_test_eq(cpl_mask_count(collapsed), ny - nzero);

                cpl_mask_delete(collapsed);


                ttemp = cpl_test_get_cputime();

                collapsed = cpl_mask_collapse_create(raw, 0);
                tcol0 += cpl_test_get_cputime() - ttemp;

                cpl_test_error(CPL_ERROR_NONE);
                cpl_test_nonnull(collapsed);

                cpl_test_eq(cpl_mask_get_size_x(collapsed), nx);
                cpl_test_eq(cpl_mask_get_size_y(collapsed), 1);

                for (i = 1; i <= nmin; i += 2) {
                    cpl_test_zero(cpl_mask_get(collapsed, i, 1));
                }

                cpl_test_eq(cpl_mask_count(collapsed), nx - nzero);

                cpl_mask_delete(collapsed);


                /* Fill the mask */
                ttemp = cpl_test_get_cputime();
                cpl_test_zero(cpl_mask_xor(raw, raw));
                cpl_test(cpl_mask_is_empty(raw));
                cpl_test_zero(cpl_mask_not(raw));
                tassi += cpl_test_get_cputime() - ttemp;
                cpl_test_eq(cpl_mask_count(raw), nx * ny);

                /* The first row is all CPL_BINARY_0 */
                for (i = 1; i <= nx; i++) {
                    cpl_test_zero(cpl_mask_set(raw, i, 1, CPL_BINARY_0));
                }
                cpl_test_eq(cpl_mask_count(raw), nx * (ny-1));

                ttemp = cpl_test_get_cputime();

                collapsed = cpl_mask_collapse_create(raw, 0);
                tcol0 += cpl_test_get_cputime() - ttemp;

                cpl_test_error(CPL_ERROR_NONE);
                cpl_test_nonnull(collapsed);

                cpl_test_eq(cpl_mask_get_size_x(collapsed), nx);
                cpl_test_eq(cpl_mask_get_size_y(collapsed), 1);

                cpl_test(cpl_mask_is_empty(collapsed));

                cpl_mask_delete(collapsed);

            }

            cpl_mask_delete(ref);
            cpl_mask_delete(raw);
        }
    }

    cpl_msg_info(cpl_func, "Time to flip [s]: %g", tflip);
    cpl_msg_info(cpl_func, "Time to turn [s]: %g", tturn);
    cpl_msg_info(cpl_func, "Time to assign [s]: %g", tassi);
    cpl_msg_info(cpl_func, "Time to collapse 0 [s]: %g", tcol0);
    cpl_msg_info(cpl_func, "Time to collapse 1 [s]: %g", tcol1);

    return;
}

